# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
import os
import platform
import shutil
from unittest import mock

import pytest
from callee import Contains, Matching, StartsWith

from mozphab import mozphab

from .conftest import hg_out, write_text

# Fail if arc ping is called
arc_ping = mock.Mock()
arc_ping.return_value = False

call_conduit = mock.Mock()


def test_submit_create(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    a_rename = hg_repo_path / "A to rename"
    write_text(a_rename, "rename me\nsecond line\n")

    b_remove = hg_repo_path / "B to remove"
    write_text(b_remove, "remove me\n")

    c_modify = hg_repo_path / "C to modify"
    write_text(c_modify, "modify me\n")

    d_copy = hg_repo_path / "D to copy"
    write_text(d_copy, "copy me\n")

    hg_out("add")
    hg_out("commit", "-m", "first")

    a_rename.rename(hg_repo_path / "A renamed")
    b_remove.unlink()
    write_text(c_modify, "modified\n")
    hg_out("copy", d_copy, "D copied")

    e_subdir = hg_repo_path / "subdir"
    e_subdir.mkdir()
    write_text(e_subdir / "E add", "added\n")

    hg_out("addremove")

    msg = hg_repo_path / "msg"
    write_text(msg, "Ą r?alice", encoding="utf-8")
    hg_out("commit", "-l", msg)
    msg.unlink()

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - Ą r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert mock.call("conduit.ping", {}) in call_conduit.call_args_list
    assert (
        mock.call("user.query", {"usernames": ["alice"]}) in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "diffusion.repository.search",
            {"limit": 1, "constraints": {"callsigns": ["TEST"]}},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {"diff_id": "1", "name": "local:commits", "data": Contains('"rev": "')},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "hg",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "default",
                "changes": [
                    {
                        "currentPath": "A renamed",
                        "type": 6,  # MOVE_HERE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " rename me\n second line\n",
                            }
                        ],
                        "oldProperties": {},
                        "oldPath": "A to rename",
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "A to rename",
                        "type": 4,  # MOVE_AWAY
                        "hunks": [],
                        "oldProperties": {},
                        "oldPath": None,
                        "commitHash": mock.ANY,
                        "awayPaths": ["A renamed"],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "B to remove",
                        "type": 3,  # DELETE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 0,
                                "newLength": 0,
                                "addLines": 0,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-remove me\n",
                            }
                        ],
                        "awayPaths": [],
                        "fileType": 1,
                        "oldPath": "B to remove",
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "oldProperties": {"unix:filemode": "100644"},
                    },
                    {
                        "currentPath": "C to modify",
                        "type": 2,  # CHANGE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-modify me\n+modified\n",
                            }
                        ],
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "fileType": 1,
                        "oldPath": "C to modify",
                        "newProperties": {},
                        "awayPaths": [],
                        "oldProperties": {},
                    },
                    {
                        "metadata": {},
                        "oldPath": "D to copy",
                        "currentPath": "D copied",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,  # COPY HERE
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " copy me\n",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "D to copy",
                        "awayPaths": ["D copied"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "currentPath": "subdir/E add",
                        "type": 1,  # ADD
                        "hunks": [
                            {
                                "corpus": "+added\n",
                                "addLines": 1,
                                "oldOffset": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "oldLength": 0,
                                "isMissingNewNewline": False,
                            }
                        ],
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "oldProperties": {},
                        "fileType": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-hg",
            },
        )
        in call_conduit.call_args_list
    )
    call_conduit.assert_any_call(
        "differential.revision.edit",
        {
            "transactions": [
                {"type": "title", "value": StartsWith("Bug 1 - ")},
                {"type": "summary", "value": ""},
                {"type": "reviewers.set", "value": ["PHID-USER-1"]},
                {"type": "bugzilla.bug-id", "value": "1"},
                {"type": "update", "value": "PHID-DIFF-1"},
            ]
        },
    )


def test_submit_create_no_trailing_newline(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    a_rename = hg_repo_path / "A to rename"
    write_text(a_rename, "rename me\nsecond line")

    b_remove = hg_repo_path / "B to remove"
    write_text(b_remove, "remove me")

    c_modify = hg_repo_path / "C to modify"
    write_text(c_modify, "modify me")

    d_copy = hg_repo_path / "D to copy"
    write_text(d_copy, "copy me")

    hg_out("add")
    hg_out("commit", "-m", "first")

    a_rename.rename(hg_repo_path / "A renamed")
    b_remove.unlink()
    write_text(c_modify, "modified")
    hg_out("copy", d_copy, "D copied")

    e_subdir = hg_repo_path / "subdir"
    e_subdir.mkdir()
    write_text(e_subdir / "E add", "added")

    hg_out("addremove")

    msg = hg_repo_path / "msg"
    write_text(msg, "Ą r?alice", encoding="utf-8")
    hg_out("commit", "-l", msg)
    msg.unlink()

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - Ą r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert mock.call("conduit.ping", {}) in call_conduit.call_args_list
    assert (
        mock.call("user.query", {"usernames": ["alice"]}) in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "diffusion.repository.search",
            {"limit": 1, "constraints": {"callsigns": ["TEST"]}},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.setdiffproperty",
            {"diff_id": "1", "name": "local:commits", "data": Contains('"rev": "')},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "hg",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "default",
                "changes": [
                    {
                        "currentPath": "A renamed",
                        "type": 6,  # MOVE_HERE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " rename me\n second line",
                            }
                        ],
                        "oldProperties": {},
                        "oldPath": "A to rename",
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "A to rename",
                        "type": 4,  # MOVE_AWAY
                        "hunks": [],
                        "oldProperties": {},
                        "oldPath": None,
                        "commitHash": mock.ANY,
                        "awayPaths": ["A renamed"],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "B to remove",
                        "type": 3,  # DELETE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 0,
                                "newLength": 0,
                                "addLines": 0,
                                "delLines": 1,
                                "isMissingOldNewline": True,
                                "isMissingNewNewline": False,
                                "corpus": "-remove me\n\\ No newline at end of file\n",
                            }
                        ],
                        "awayPaths": [],
                        "fileType": 1,
                        "oldPath": "B to remove",
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "oldProperties": {"unix:filemode": "100644"},
                    },
                    {
                        "currentPath": "C to modify",
                        "type": 2,  # CHANGE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": True,
                                "isMissingNewNewline": True,
                                "corpus": (
                                    "-modify me\n"
                                    "\\ No newline at end of file\n"
                                    "+modified\n"
                                    "\\ No newline at end of file\n"
                                ),
                            }
                        ],
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "fileType": 1,
                        "oldPath": "C to modify",
                        "newProperties": {},
                        "awayPaths": [],
                        "oldProperties": {},
                    },
                    {
                        "metadata": {},
                        "oldPath": "D to copy",
                        "currentPath": "D copied",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,  # COPY HERE
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " copy me",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "D to copy",
                        "awayPaths": ["D copied"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "currentPath": "subdir/E add",
                        "type": 1,  # ADD
                        "hunks": [
                            {
                                "oldOffset": 0,
                                "oldLength": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": True,
                                "corpus": "+added\n\\ No newline at end of file\n",
                            }
                        ],
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "oldProperties": {},
                        "fileType": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-hg",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_create_no_bug(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    write_text(hg_repo_path / "x", "a")
    hg_out("add")
    hg_out("commit", "--message", "A r?alice")
    mozphab.main(["submit", "--yes", "--no-bug", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
A r?alice

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()


def test_submit_create_binary(in_process, hg_repo_path, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # file.allocate
        {"filePHID": None, "upload": True},
        # file.upload
        {},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    shutil.copyfile(str(data_file), str(hg_repo_path / "img.png"))
    hg_out("add")
    hg_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.upload", {"data_base64": mock.ANY, "name": "img.png"})
        in call_conduit.call_args_list
    )


def test_submit_create_binary_existing(in_process, hg_repo_path, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # file.allocate
        {"filePHID": "PHID-FILE-1", "upload": False},
        # no file.upload call
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    shutil.copyfile(str(data_file), str(hg_repo_path / "img.png"))
    hg_out("add")
    hg_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert mock.call("file.upload", mock.ANY) not in call_conduit.call_args_list


def test_submit_create_binary_chunked(in_process, hg_repo_path, data_file):
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # file.allocate
        {"filePHID": "PHID-FILE-1", "upload": True},
        # file.querychunks
        [
            {"byteStart": "0", "byteEnd": "4194304", "complete": False},
            {"byteStart": "4194304", "byteEnd": "8388608", "complete": False},
            {"byteStart": "8388608", "byteEnd": "8425160", "complete": False},
        ],
        # file.uploadchunk
        {},
        # file.uploadchunk
        {},
        # file.uploadchunk
        {},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    shutil.copyfile(str(data_file), str(hg_repo_path / "img.png"))
    hg_out("add")
    hg_out("commit", "-m", "IMG")

    mozphab.main(["submit", "--yes", "--bug", "1", "."], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - IMG

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "file.allocate",
            {"name": "img.png", "contentHash": mock.ANY, "contentLength": 182},
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call("file.querychunks", {"filePHID": "PHID-FILE-1"})
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 0,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 4194304,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )
    assert (
        mock.call(
            "file.uploadchunk",
            {
                "filePHID": "PHID-FILE-1",
                "byteStart": 8388608,
                "data": mock.ANY,
                "dataEncoding": "base64",
            },
        )
        in call_conduit.call_args_list
    )


def test_submit_create_no_checkout(in_process, hg_repo_path):
    """Test that diffing behaviour is consistent even without an explicit
    checkout prior to generating diffs.

    This is a counterpart to test_integration_git.test_submit_create_no_checkout
    and is named the same for ease of localisation. It doesn't however, test
    that the Hg logic doesn't do checkouts during submissions. Currently, it
    very much does, and diffs for files touched by multiple commits end up
    containing, incorrectly, all the changes up to the tip of the branche (see
    bug 1926924).
    """
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # First diff
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
        # Second diff
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "124", "phid": "PHID-DREV-124"}},
        # differential.setdiffproperty
        {},
    )
    a_rename = hg_repo_path / "A to rename"
    write_text(a_rename, "rename me\nsecond line\n")

    b_remove = hg_repo_path / "B to remove"
    write_text(b_remove, "remove me\n")

    c_modify = hg_repo_path / "C to modify"
    write_text(c_modify, "modify me\n")

    d_copy = hg_repo_path / "D to copy"
    write_text(d_copy, "copy me\n")

    hg_out("add")
    hg_out("commit", "-m", "first")

    a_rename.rename(hg_repo_path / "A renamed")
    b_remove.unlink()
    write_text(c_modify, "modified\n")
    hg_out("copy", d_copy, "D copied")

    e_subdir = hg_repo_path / "subdir"
    e_subdir.mkdir()
    write_text(e_subdir / "E add", "added\n")

    hg_out("addremove")

    msg = hg_repo_path / "msg"
    write_text(msg, "Ą r?alice", encoding="utf-8")
    hg_out("commit", "-l", msg)
    msg.unlink()

    write_text(c_modify, "leave me alone\n")
    hg_out("add")
    hg_out("commit", "-m", "second")

    mozphab.main(["submit", "--yes", "--bug", "1", "-2"], is_development=True)

    expected_calls = [
        (
            "differential.creatediff",
            {
                "sourceControlPath": "/",
                "sourceControlSystem": "hg",
                "lintStatus": "none",
                "sourcePath": mock.ANY,
                "unitStatus": "none",
                "sourceMachine": "http://example.test",
                "sourceControlBaseRevision": mock.ANY,
                "repositoryPHID": "PHID-REPO-1",
                "branch": "default",
                "changes": [
                    {
                        "currentPath": "A renamed",
                        "type": 6,  # MOVE_HERE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 2,
                                "newOffset": 1,
                                "newLength": 2,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " rename me\n second line\n",
                            }
                        ],
                        "oldProperties": {},
                        "oldPath": "A to rename",
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "A to rename",
                        "type": 4,  # MOVE_AWAY
                        "hunks": [],
                        "oldProperties": {},
                        "oldPath": None,
                        "commitHash": mock.ANY,
                        "awayPaths": ["A renamed"],
                        "metadata": {},
                        "newProperties": {},
                        "fileType": 1,
                    },
                    {
                        "currentPath": "B to remove",
                        "type": 3,  # DELETE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 0,
                                "newLength": 0,
                                "addLines": 0,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-remove me\n",
                            }
                        ],
                        "awayPaths": [],
                        "fileType": 1,
                        "oldPath": "B to remove",
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "oldProperties": {"unix:filemode": "100644"},
                    },
                    {
                        "currentPath": "C to modify",
                        "type": 2,  # CHANGE
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-modify me\n+modified\n",
                            }
                        ],
                        "commitHash": mock.ANY,
                        "metadata": {},
                        "fileType": 1,
                        "oldPath": "C to modify",
                        "newProperties": {},
                        "awayPaths": [],
                        "oldProperties": {},
                    },
                    {
                        "metadata": {},
                        "oldPath": "D to copy",
                        "currentPath": "D copied",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,  # COPY HERE
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " copy me\n",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "D to copy",
                        "awayPaths": ["D copied"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "currentPath": "subdir/E add",
                        "type": 1,  # ADD
                        "hunks": [
                            {
                                "corpus": "+added\n",
                                "addLines": 1,
                                "oldOffset": 0,
                                "newOffset": 1,
                                "newLength": 1,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "oldLength": 0,
                                "isMissingNewNewline": False,
                            }
                        ],
                        "commitHash": mock.ANY,
                        "awayPaths": [],
                        "newProperties": {"unix:filemode": "100644"},
                        "oldPath": None,
                        "oldProperties": {},
                        "fileType": 1,
                        "metadata": {},
                    },
                ],
                "creationMethod": "moz-phab-hg",
            },
        ),
        (
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "C to modify",
                        "currentPath": "C to modify",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 1,
                                "delLines": 1,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": "-modified\n+leave me alone\n",
                            }
                        ],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        ),
    ]

    # Match expectations to the specific call to compare them to.
    # This makes failure reporting more helpful for debugging.
    matched_calls = list(
        zip(
            expected_calls,
            [
                c[0]
                for c in call_conduit.call_args_list
                if c[0][0] == "differential.creatediff"
            ],
        )
    )
    # Make sure we've got something to assert!
    assert matched_calls, "There are no differential.creatediff calls to compare!"
    for expected, actual in matched_calls:
        assert actual == expected


def test_submit_remove_cr(in_process, hg_repo_path):
    call_conduit.reset_mock()
    call_conduit.side_effect = [
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # user.search
        [{"userName": "alice", "phid": "PHID-USER-1"}],
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    ]
    test_a = hg_repo_path / "X"
    with open(test_a, "w", newline="") as f:
        f.write("a\r\nb\n")
    hg_out("add", "X")
    hg_out("commit", "--message", "A r?alice")
    # removing CR, leaving LF
    with open(test_a, "w", newline="") as f:
        f.write("a\nb\n")
    hg_out("commit", "--message", "B r?alice")

    mozphab.main(["submit", "--yes", "--bug", "1", "-s"], is_development=True)

    call_conduit.assert_any_call(
        "differential.creatediff",
        Matching(lambda x: x["changes"][0]["hunks"][0]["corpus"] == "-a\r\n+a\n b\n"),
    )


def test_submit_single_first(in_process, hg_repo_path, hg_sha):
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    write_text(hg_repo_path / "X", "a\n")
    hg_out("add", "X")
    hg_out("commit", "-m", "A")
    sha = hg_sha()
    write_text(hg_repo_path / "X", "b\n")
    hg_out("commit", "-m", "B")

    mozphab.main(
        ["submit", "--yes", "--bug", "1", "--single", sha], is_development=True
    )

    log = hg_out("log", "--template", r"{desc}\n---\n", "--limit", "2")
    expected = """\
B
---
Bug 1 - A

Differential Revision: http://example.test/D123
---
"""
    assert log == expected


def test_submit_single_last(in_process, hg_repo_path):
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    write_text(hg_repo_path / "X", "a\n")
    hg_out("add", "X")
    hg_out("commit", "-m", "A")
    write_text(hg_repo_path / "X", "b\n")
    hg_out("commit", "-m", "B")

    mozphab.main(["submit", "--yes", "--bug", "1", "--single"], is_development=True)

    log = hg_out("log", "--template", r"{desc}\n---\n", "--limit", "2")
    expected = """\
Bug 1 - B

Differential Revision: http://example.test/D123
---
A
---
"""
    assert log == expected


def test_multiple_copy(in_process, hg_repo_path):
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    write_text(hg_repo_path / "X", "a\n")
    hg_out("add", "X")
    hg_out("commit", "-m", "A")
    hg_out("cp", "X", "X_copy_1")
    hg_out("cp", "X", "X_copy_2")
    hg_out("commit", "-m", "multiple copy")

    mozphab.main(["submit", "--yes", "--bug", "1", "--single"], is_development=True)

    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "X",
                        "awayPaths": ["X_copy_1", "X_copy_2"],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 5,  # COPY_AWAY
                        "fileType": 1,
                        "hunks": [],
                    },
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X_copy_1",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " a\n",
                            }
                        ],
                    },
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X_copy_2",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 7,
                        "fileType": 1,
                        "hunks": [
                            {
                                "oldOffset": 1,
                                "oldLength": 1,
                                "newOffset": 1,
                                "newLength": 1,
                                "addLines": 0,
                                "delLines": 0,
                                "isMissingOldNewline": False,
                                "isMissingNewNewline": False,
                                "corpus": " a\n",
                            }
                        ],
                    },
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    )


# To re-enable this test for Windows we would need a way to modify
# UMASK bits similar to how it's done in test_integration_git.py.
# However, Mercurial doesn't have an equivalent to `git update-index`.
@pytest.mark.skipif(
    platform.system() == "Windows",
    reason="Skipped because Windows cannot modify UMASK bits",
)
def test_empty_file(in_process, hg_repo_path, hg_sha):
    # Add empty file
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # ping
        {},
        # diffusion.repository.search
        {"data": [{"phid": "PHID-REPO-1", "fields": {"vcs": "hg"}}]},
        # differential.creatediff
        {"phid": "PHID-DIFF-1", "diffid": "1"},
        # differential.revision.edit
        {"object": {"id": "123", "phid": "PHID-DREV-123"}},
        # differential.setdiffproperty
        {},
    )
    testfile = hg_repo_path / "X"
    testfile.touch()
    hg_out("add")
    hg_out("commit", "-m", "A")
    sha = hg_sha()
    mozphab.main(["submit", "--yes", "--bug", "1", sha], is_development=True)
    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - A

Differential Revision: http://example.test/D123
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": None,
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {},
                        "newProperties": {"unix:filemode": "100644"},
                        "commitHash": mock.ANY,
                        "type": 1,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have populated fields and "
        "`change.hunks` should be an empty list"
    )

    # Modify a file's mode
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        {"phid": "PHID-DIFF-2", "diffid": "2"},
        # differential.revision.edit
        {"object": {"id": "124", "phid": "PHID-DREV-124"}},
        # differential.setdiffproperty
        {},
    )
    os.chmod(testfile, 0o0755)

    hg_out("commit", "-m", "B")
    sha = hg_sha()
    mozphab.main(["submit", "--yes", "--bug", "1", sha], is_development=True)
    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - B

Differential Revision: http://example.test/D124
"""
    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100644"},
                        "newProperties": {"unix:filemode": "100755"},
                        "commitHash": mock.ANY,
                        "type": 2,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should contain filemode changes and "
        "`change.hunks` should be an empty list"
    )

    # Remove an empty file
    call_conduit.reset_mock()
    call_conduit.side_effect = (
        # differential.creatediff
        {"phid": "PHID-DIFF-3", "diffid": "3"},
        # differential.revision.edit
        {"object": {"id": "125", "phid": "PHID-DREV-125"}},
        # differential.setdiffproperty
        {},
    )
    testfile.unlink()
    hg_out("addremove")
    hg_out("commit", "-m", "C")
    sha = hg_sha()
    mozphab.main(["submit", "--yes", "--bug", "1", sha], is_development=True)
    log = hg_out("log", "--template", r"{desc}\n", "--rev", ".")
    expected = """
Bug 1 - C

Differential Revision: http://example.test/D125
"""

    assert log.strip() == expected.strip()
    assert (
        mock.call(
            "differential.creatediff",
            {
                "changes": [
                    {
                        "metadata": {},
                        "oldPath": "X",
                        "currentPath": "X",
                        "awayPaths": [],
                        "oldProperties": {"unix:filemode": "100755"},
                        "newProperties": {},
                        "commitHash": mock.ANY,
                        "type": 3,
                        "fileType": 1,
                        "hunks": [],
                    }
                ],
                "sourceMachine": "http://example.test",
                "sourceControlSystem": "hg",
                "sourceControlPath": "/",
                "sourceControlBaseRevision": mock.ANY,
                "creationMethod": "moz-phab-hg",
                "lintStatus": "none",
                "unitStatus": "none",
                "repositoryPHID": "PHID-REPO-1",
                "sourcePath": mock.ANY,
                "branch": "default",
            },
        )
        in call_conduit.call_args_list
    ), (
        "The diff should have empty `newProperties` and "
        "`change.hunks` should be an empty list"
    )



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import subprocess
from pathlib import Path

from .conftest import find_script_path

ROOT = Path(__file__).resolve().parent.parent
PY_FILES = sorted(
    str(f)
    for f in list(ROOT.glob("*.py"))
    + list((ROOT / "mozphab").glob("**/*.py"))
    + list((ROOT / "tests").glob("**/*.py"))
)


def test_black():
    subprocess.check_call([find_script_path("black"), "--check"] + PY_FILES)


def test_ruff():
    """Run ruff on the codebase.

    Use the project root as the directory to lint, and define appropriate lint
    paths in the `ruff.toml` file.
    """
    subprocess.check_call((find_script_path("ruff"), "check", ROOT))



# coding=utf-8
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import copy
import unittest
import uuid
from unittest import mock

from callee import Contains

from mozphab import environment, exceptions, helpers, mozphab, repository
from mozphab.commands import submit
from mozphab.commits import Commit

from .conftest import search_diff, search_rev


def reviewers_dict(reviewers=None):
    return {
        "request": reviewers[0] if reviewers else [],
        "granted": reviewers[1] if reviewers else [],
    }


def commit(
    bug_id=None,
    reviewers=None,
    rev_id=None,
    node=None,
    title="",
    body="",
    wip=False,
    bug_id_orig=None,
    submit=True,
):
    node = node or f"{uuid.uuid4().hex}12345678"
    return Commit(
        name=helpers.short_node(node),
        title=title,
        title_preview=title,
        bug_id=bug_id,
        bug_id_orig=bug_id_orig,
        reviewers=reviewers_dict(reviewers),
        body=body,
        rev_id=rev_id,
        node=node,
        submit=submit,
        wip=wip,
    )


class Args(argparse.Namespace):
    def __init__(
        self,
        command="submit",
        reviewer=None,
        blocker=None,
        bug=None,
        wip=False,
        no_wip=False,
        no_bug=False,
        force=False,
        single=False,
        end_rev=environment.DEFAULT_END_REV,
    ):
        self.command = command
        self.reviewer = reviewer
        self.blocker = blocker
        self.bug = bug
        self.wip = wip
        self.no_wip = no_wip
        self.no_bug = no_bug
        self.force = force
        self.single = single
        self.end_rev = end_rev


# noinspection PyPep8Naming,PyBroadException
class Commits(unittest.TestCase):
    def _assertNoError(self, callableObj, *args, **kwargs):
        try:
            callableObj(*args, **kwargs)
        except exceptions.Error as e:
            self.fail("%s raised" % e)

    def _assertError(self, callableObj, expected, *args, **kwargs):
        with self.assertRaisesRegex(exceptions.Error, expected):
            callableObj(*args, **kwargs)

    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    @mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
    def test_commit_stack_validation_errors(self, m_revs, check_reviewers):
        check_reviewers.return_value = []
        m_revs.return_value = []

        _, errors = submit.validate_commit_stack([commit("1", (["r"], []))], Args())
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack(
            [
                commit("1", (["r1"], [])),
                commit("2", (["r1"], [])),
                commit("3", (["r1", "r2"], [])),
            ],
            Args(),
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack([commit("1", None)], Args())
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack(
            [commit("1", (["r"], [])), commit("1", None)], Args()
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack([commit(None, (["r"], []))], Args())
        self.assertEqual(list(errors.values()), [[Contains("Missing bug ID")]])

        _, errors = submit.validate_commit_stack(
            [commit(None, (["r"], []))], Args(no_bug=True)
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack([commit("", (["r"], []))], Args())
        self.assertEqual(list(errors.values()), [[Contains("Missing bug ID")]])

        _, errors = submit.validate_commit_stack(
            [commit("1", (["r"], [])), commit("", (["r"], []))], Args()
        )
        self.assertEqual(list(errors.values()), [[Contains("Missing bug ID")]])

        _, errors = submit.validate_commit_stack(
            [commit("1", (["r"], [])), commit("", (["r"], []))], Args(no_bug=True)
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack(
            [commit("1", (["r"], []), body="Summary: blah\nReviewers: r")], Args()
        )
        self.assertEqual(list(errors.values()), [[Contains("Contains arc fields")]])

        _, errors = submit.validate_commit_stack([commit(bug_id="1", rev_id=1)], Args())
        self.assertEqual(
            list(errors.values()),
            [[Contains("didn't return a query result for revision D1")]],
        )

    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    @mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
    @mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
    @mock.patch("mozphab.conduit.ConduitAPI.whoami")
    def test_invalid_reviewers_fails_the_stack_validation_check(
        self, m_whoami, m_diffs, m_revs, check_reviewers
    ):
        def fail_gonzo(reviewers):
            # Replace the check_for_invalid_reviewers() function with something that
            # fails for certain reviewers.
            if "gonzo" in reviewers["request"]:
                return [{"name": "gonzo"}]
            elif "goober" in reviewers["request"]:
                return [{"name": "goober", "until": "string"}]
            elif "goofus" in reviewers["request"]:
                return [{"name": "goofus", "disabled": True}]
            else:
                return []

        check_reviewers.side_effect = fail_gonzo
        m_revs.return_value = []
        m_diffs.return_value = {"PHID-DIFF-1": search_diff()}
        m_whoami.return_value = {"phid": "PHID-USER-1"}

        _, errors = submit.validate_commit_stack(
            # Build a stack with an invalid reviewer in the middle.
            [
                commit("1", (["alice"], [])),
                commit("2", (["bob", "gonzo"], [])),
                commit("3", (["charlie"], [])),
            ],
            Args(),
        )
        self.assertEqual(
            list(errors.values()), [[Contains("gonzo isn't a valid reviewer")]]
        )

        _, errors = submit.validate_commit_stack(
            # Build a stack with an unavailable reviewer in the middle.
            [
                commit("1", (["alice"], [])),
                commit("2", (["bob", "goober"], [])),
                commit("3", (["charlie"], [])),
            ],
            Args(),
        )
        self.assertEqual(
            list(errors.values()), [[Contains("goober isn't available until string")]]
        )

        warnings, errors = submit.validate_commit_stack(
            [
                commit("1", (["alice"], [])),
                commit("2", (["bob", "goober"], [])),
                commit("3", (["charlie"], [])),
            ],
            Args(force=True),
        )
        self.assertEqual(
            list(warnings.values()), [[Contains("goober isn't available until string")]]
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack(
            [commit("1", (["goofus"], []))], Args()
        )
        self.assertEqual(list(errors.values()), [[Contains("goofus is disabled")]])

        # Never error for an existing revision with reviewers.
        m_revs.return_value = [search_rev(reviewers=["PHID-USER-2"])]
        _, errors = submit.validate_commit_stack(
            [commit("1", (["goofus"], []), rev_id=1)], Args()
        )
        self.assertEqual(errors, {})

    @mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
    @mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
    @mock.patch("mozphab.conduit.ConduitAPI.whoami")
    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    def test_validate_duplicate_revision(
        self, check_reviewers, m_whoami, m_diffs, m_revs
    ):
        check_reviewers.return_value = []
        m_revs.return_value = [
            search_rev(rev=1),
            search_rev(rev=2),
            search_rev(rev=3),
        ]
        m_diffs.return_value = {"PHID-DIFF-1": search_diff()}
        m_whoami.return_value = {"phid": "PHID-USER-1"}

        _, errors = submit.validate_commit_stack(
            [
                commit("1", (["r"], []), node="a"),
                commit("2", (["r"], []), node="b"),
                commit("3", (["r"], []), node="c"),
            ],
            Args(),
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack(
            [
                commit("1", (["r"], []), rev_id=1, node="a"),
                commit("2", (["r"], []), rev_id=2, node="b"),
                commit("3", (["r"], []), rev_id=3, node="c"),
            ],
            Args(),
        )
        self.assertEqual(errors, {})

        _, errors = submit.validate_commit_stack(
            [
                commit("1", (["r"], []), rev_id=1, node="a"),
                commit("2", (["r"], []), rev_id=2, node="b"),
                commit("3", (["r"], []), rev_id=1, node="c"),
            ],
            Args(),
        )
        self.assertEqual(
            errors, {"c": [Contains("commit a refers to the same one D1")]}
        )

        _, errors = submit.validate_commit_stack(
            [
                commit("1", (["r"], []), rev_id=1, node="a"),
                commit("2", (["r"], []), rev_id=2, node="b"),
                commit("3", (["r"], []), rev_id=1, node="c"),
                commit("4", (["r"], []), rev_id=2, node="d"),
            ],
            Args(),
        )
        self.assertEqual(
            errors,
            {
                "c": [Contains("commit a refers to the same one D1")],
                "d": [Contains("commit b refers to the same one D2")],
            },
        )

    def test_commit_preview(self):
        build = helpers.build_commit_title

        self.assertEqual(
            "Bug 1, blah, r=turnip",
            build(commit("1", ([], ["turnip"]), title="bug 1, blah, r=turnip")),
        )
        self.assertEqual(
            "blah (Bug 1) r=turnip",
            build(commit("1", ([], ["turnip"]), title="blah (bug 1) r=turnip")),
        )
        self.assertEqual(
            "Bug 1 - blah r?turnip",
            build(commit("1", (["turnip"], []), title="blah r?turnip")),
        )

        self.assertEqual(
            "blah r=turnip", build(commit("", ([], ["turnip"]), title="blah r=turnip"))
        )
        self.assertEqual(
            "Bug 1 - blah", build(commit("1", None, title="Bug 1 - blah r?turnip"))
        )
        self.assertEqual(
            "Bug 1 - blah", build(commit("1", None, title="Bug 1 - blah r=turnip"))
        )
        self.assertEqual(
            "Bug 1 - helper_bug2.html",
            build(commit("1", None, title="Bug 1 - helper_bug2.html")),
        )

    @mock.patch("mozphab.helpers.build_commit_title")
    def test_update_commit_title_previews(self, m_build_commit_title):
        m_build_commit_title.side_effect = lambda x: x.title + " preview"
        commits = [Commit(title="a"), Commit(title="b")]
        helpers.update_commit_title_previews(commits)
        self.assertEqual(
            [
                Commit(title="a", title_preview="a preview"),
                Commit(title="b", title_preview="b preview"),
            ],
            commits,
        )

    def test_replace_request_reviewers(self):
        replace = helpers.replace_reviewers
        self.assertEqual("", replace("", reviewers_dict()))
        self.assertEqual("Title", replace("Title", reviewers_dict()))
        self.assertEqual(
            "Title\n\nr?one r=two", replace("Title\n\nr?one r=two", reviewers_dict())
        )
        self.assertEqual("r?one", replace("", reviewers_dict([["one"], []])))
        self.assertEqual("r?one", replace("r?one", reviewers_dict([["one"], []])))
        self.assertEqual("r?one,two", replace("", reviewers_dict([["one", "two"], []])))
        self.assertEqual(
            "Some Title r?one,two",
            replace("Some Title", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one\n\nDescr\niption",
            replace("Title\n\nDescr\niption", reviewers_dict([["one"], []])),
        )
        self.assertEqual(
            "Title r?one,two\n\nr?two",
            replace("Title\n\nr?two", reviewers_dict([["one", "two"], []])),
        )

        self.assertEqual("Title", replace("Title r?one", reviewers_dict()))
        self.assertEqual("Title", replace("Title r?one,two", reviewers_dict()))
        self.assertEqual("Title", replace("Title r?one r?two", reviewers_dict()))
        self.assertEqual(
            "Title r?one!", replace("Title r?one!", reviewers_dict([["one!"], []]))
        )
        self.assertEqual(
            "Title r?one", replace("Title r?one!", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one!", replace("Title r?one", reviewers_dict([["one!"], []]))
        )
        self.assertEqual(
            "Title r?one", replace("Title r?one", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one one", replace("Title r? one", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one,two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one r?two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one", replace("Title r=one", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one", replace("Title r=one,two", reviewers_dict([["one"], []]))
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r=one,two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r=one r=two", reviewers_dict([["one", "two"], []])),
        )
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one r=two", reviewers_dict([["one", "two"], []])),
        )

    def test_replace_granted_reviewers(self):
        replace = helpers.replace_reviewers
        self.assertEqual("r=one", replace("", reviewers_dict([[], ["one"]])))
        self.assertEqual("r=one", replace("r=one", reviewers_dict([[], ["one"]])))
        self.assertEqual("r=one,two", replace("", reviewers_dict([[], ["one", "two"]])))
        self.assertEqual(
            "Some Title r=one,two",
            replace("Some Title", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one\n\nDescr\niption",
            replace("Title\n\nDescr\niption", reviewers_dict([[], ["one"]])),
        )
        self.assertEqual(
            "Title r=one,two\n\nr?two",
            replace("Title\n\nr?two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual("Title", replace("Title r=one", reviewers_dict()))
        self.assertEqual("Title", replace("Title r=one,two", reviewers_dict()))
        self.assertEqual("Title", replace("Title r=one r=two", reviewers_dict()))
        self.assertEqual(
            "Title r=one", replace("Title r=one", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one!", replace("Title r=one!", reviewers_dict([[], ["one!"]]))
        )
        self.assertEqual(
            "Title r=one", replace("Title r=one!", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one!", replace("Title r=one", reviewers_dict([[], ["one!"]]))
        )
        self.assertEqual(
            "Title r=one one", replace("Title r= one", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=one,two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=one r=two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one", replace("Title r?one", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one", replace("Title r?one,two", reviewers_dict([[], ["one"]]))
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r?one,two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r?one r?two", reviewers_dict([[], ["one", "two"]])),
        )
        self.assertEqual(
            "Title r=one,two",
            replace("Title r=one r?two", reviewers_dict([[], ["one", "two"]])),
        )

    def test_replace_mixed_reviewers(self):
        replace = helpers.replace_reviewers
        self.assertEqual(
            "Title r?one r=two", replace("Title", reviewers_dict([["one"], ["two"]]))
        )
        self.assertEqual(
            "Title r?one r=two",
            replace("Title r=one r?two", reviewers_dict([["one"], ["two"]])),
        )
        self.assertEqual(
            "Title r?one r=two",
            replace("Title r?two r=one", reviewers_dict([["one"], ["two"]])),
        )
        self.assertEqual(
            "Title r?one,two r=three",
            replace("Title r=one r?two", reviewers_dict([["one", "two"], ["three"]])),
        )
        self.assertEqual(
            "Title r?one r=two,three",
            replace("Title r=one r?two", reviewers_dict([["one"], ["two", "three"]])),
        )

    @unittest.skip("These tests should pass we should fix the function")
    def test_badly_replaced_reviewers(self):
        replace = mozphab.replace_reviewers
        # r?two
        self.assertEqual("r?one", replace("r?two", reviewers_dict([["one"], []])))
        # r=one
        self.assertEqual("r?one", replace("r=one", reviewers_dict([["one"], []])))
        # r? one
        self.assertEqual("r?one one", replace("r? one", reviewers_dict([["one"], []])))
        # r?one,
        self.assertEqual("r?one", replace("r?one,", reviewers_dict([["one"], []])))
        # r?one
        self.assertEqual(
            "r?one,two", replace("r?one", reviewers_dict([["one", "two"], []]))
        )
        # r?one,two
        self.assertEqual("r?one", replace("r?one,two", reviewers_dict([["one"], []])))
        # Title r?one,two,,two
        self.assertEqual(
            "Title r?one,two",
            replace("Title r?one,,two", reviewers_dict([["one"], ["two"], []])),
        )
        # r?one
        self.assertEqual("", replace("r?one", reviewers_dict()))
        # r?one,two
        self.assertEqual("", replace("r?one,two", reviewers_dict()))
        # r?one
        self.assertEqual("", replace("r?one r?two", reviewers_dict()))
        # r?two
        self.assertEqual(
            "r?one,two", replace("r?two", reviewers_dict([["one", "two"], []]))
        )
        # r?one r?one,two
        self.assertEqual(
            "r?one,two", replace("r?one r?two", reviewers_dict([["one", "two"], []]))
        )
        # r=one
        self.assertEqual("r?one", replace("r=one", reviewers_dict([["one"], []])))
        # r=one,two
        self.assertEqual("r?one", replace("r=one,two", reviewers_dict([["one"], []])))
        # r=one,two
        self.assertEqual(
            "r?one,two", replace("r=one,two", reviewers_dict([["one", "two"], []]))
        )
        # r=one, r?one,two
        self.assertEqual(
            "r?one,two", replace("r=one r=two", reviewers_dict([["one", "two"], []]))
        )
        # r? one, r?one,two
        self.assertEqual(
            "r?one,two", replace("r?one r=two", reviewers_dict([["one", "two"], []]))
        )

        # Granted
        # r=two
        self.assertEqual("r=one", replace("r=two", reviewers_dict([[], ["one"]])))
        # r?one
        self.assertEqual("r=one", replace("r?one", reviewers_dict([[], ["one"]])))
        # r?one
        self.assertEqual(
            "r=one,two", replace("r=one", reviewers_dict([[], ["one", "two"]]))
        )
        # r?one,two
        self.assertEqual("r?one", replace("r?one,two", reviewers_dict([["one"], []])))

    def test_show_commit_stack(self):
        submit.conduit.set_repo(repository.Repository("", "", "http://phab"))

        with self.assertLogs() as logging_watcher:
            submit.log_commit_stack_with_messages([commit(node="aaa000", title="A")])
        self.assertEqual(logging_watcher.output, [Contains("(New) aaa000 A")])

        with self.assertLogs() as logging_watcher:
            submit.log_commit_stack_with_messages(
                [commit(rev_id=12, node="aaa000", title="A")]
            )
        self.assertEqual(logging_watcher.output, [Contains("(D12) aaa000 A")])

        with self.assertLogs() as logging_watcher:
            submit.log_commit_stack_with_messages(
                [
                    commit(node="aaa000", title="A"),
                    commit(node="bbb000", title="B"),
                ]
            )
        self.assertEqual(
            logging_watcher.output,
            [Contains("(New) bbb000 B"), Contains("(New) aaa000 A")],
        )

        with self.assertLogs() as logging_watcher:
            submit.show_commit_stack([commit(rev_id=123, node="aaa000", title="A")])
        self.assertEqual(
            logging_watcher.output,
            [Contains("(D123) aaa000 A"), Contains("-> http://phab/D123")],
        )

        # Only show commits that were submitted.
        with self.assertLogs() as logging_watcher:
            submit.show_commit_stack(
                [commit(rev_id=1, submit=False), commit(rev_id=2), commit(rev_id=3)]
            )
        self.assertEqual(
            logging_watcher.output,
            [
                Contains("(D3)"),
                Contains("-> http://phab/D3"),
                Contains("(D2)"),
                Contains("-> http://phab/D2"),
            ],
        )

    @mock.patch("mozphab.conduit.ConduitAPI.get_revisions")
    @mock.patch("mozphab.conduit.ConduitAPI.get_diffs")
    @mock.patch("mozphab.conduit.ConduitAPI.whoami")
    @mock.patch("mozphab.conduit.ConduitAPI.check_for_invalid_reviewers")
    def test_validate_commit_stack(
        self, m_check_reviewers, m_whoami, m_get_diffs, m_get_revisions
    ):
        def _commit(
            node=None,
            title="A",
            rev=None,
            bug="1",
            bug_orig=None,
            granted=None,
            request=None,
            wip=False,
        ):
            granted = granted or []
            request = request or []
            return commit(
                node=node,
                title=title,
                rev_id=rev,
                bug_id_orig=bug_orig,
                bug_id=bug,
                reviewers=(request, granted),
                wip=wip,
            )

        m_whoami.return_value = {"phid": "PHID-USER-1"}
        m_get_revisions.return_value = [search_rev()]
        m_get_diffs.return_value = {"PHID-DIFF-1": search_diff()}
        m_check_reviewers.return_value = []

        warnings, errors = submit.validate_commit_stack([], Args())
        self.assertEqual((warnings, errors), ({}, {}))

        warnings, _ = submit.validate_commit_stack(
            [_commit(bug_orig="2", bug="1", granted=["alice"])], Args()
        )
        self.assertEqual(list(warnings.values()), [["Bug ID changed from 2 to 1"]])

        # Submit with missing bug ID.
        warnings, _ = submit.validate_commit_stack(
            [_commit(bug="", request=["bob"])], Args(force=True)
        )
        self.assertEqual(list(warnings.values()), [["Missing bug ID"]])

        # Submit with missing reviewers.
        warnings, _ = submit.validate_commit_stack([_commit()], Args())
        self.assertEqual(list(warnings.values()), [["Missing reviewers"]])

        # Submit with existing revision reviewers.
        m_get_revisions.return_value = [search_rev(reviewers=["PHID-USER-2"])]
        warnings, errors = submit.validate_commit_stack([_commit(rev=1)], Args())
        self.assertEqual((warnings, errors), ({}, {}))

        # Do not update not changed commits
        m_get_revisions.return_value = [
            search_rev(),
            search_rev(rev=2, phid="PHID-REV-2", diff="PHID-DIFF-2"),
        ]
        m_get_diffs.return_value = {
            "PHID-DIFF-1": search_diff(),
            "PHID-DIFF-2": search_diff(),
        }
        # we're changing bug id in the first revision to 2
        warnings, _ = submit.validate_commit_stack(
            [
                _commit(node="aaa000aaa000", rev=1, bug="2", granted=["alice"]),
                _commit(node="bbb000bbb000", title="B", rev=2, granted=["alice"]),
            ],
            Args(),
        )
        self.assertEqual(
            warnings, {"aaa000aaa000": [Contains("revision will change from 1 to 2")]}
        )

        m_whoami.return_value = {"phid": "PHID-USER-2"}
        warnings, _ = submit.validate_commit_stack(
            [_commit(rev=1, granted=["alice"])], Args()
        )
        self.assertEqual(list(warnings.values()), [[Contains("Commandeer")]])

        # Information about not updating the commit if not changed
        m_whoami.return_value = {"phid": "PHID-USER-1"}
        m_get_revisions.return_value = [search_rev(reviewers=["PHID-USER-2"])]
        m_get_diffs.return_value = {"PHID-DIFF-1": search_diff(node="aaa000aaa000")}

        warnings, _ = submit.validate_commit_stack(
            [_commit(node="aaa000aaa000", rev=1, granted=["alice"])], Args()
        )
        self.assertEqual(
            warnings, {"aaa000aaa000": [Contains("revision has not changed")]}
        )

        # Removing the WIP state from the revision without changing the commit's SHA1
        m_get_revisions.return_value = [search_rev(status="changes-planned")]
        warnings, _ = submit.validate_commit_stack(
            [_commit(node="aaa000aaa000", rev=1, granted=["alice"])], Args()
        )
        self.assertEqual(
            warnings,
            {"aaa000aaa000": [Contains('"Changes Planned" status will change')]},
        )

        # Adding the WIP state to the revision without changing the commit's SHA1
        m_get_revisions.return_value = [search_rev()]
        warnings, _ = submit.validate_commit_stack(
            [_commit(node="aaa000aaa000", rev=1, granted=["alice"], wip=True)],
            Args(wip=True),
        )
        self.assertEqual(
            warnings,
            {"aaa000aaa000": [Contains('status will change to "Changes Planned"')]},
        )

        # Submit a new patch with reviewers.
        warnings, errors = submit.validate_commit_stack(
            [_commit(request=["alice"])], Args()
        )
        self.assertEqual((warnings, errors), ({}, {}))

        # Submit a new patch without reviewers.
        warnings, _ = submit.validate_commit_stack(
            [_commit(rev=None, request=[], wip=True)], Args()
        )
        self.assertEqual(
            list(warnings.values()),
            [
                [
                    Contains("Missing reviewers"),
                    Contains('submitted as "Changes Planned"'),
                ]
            ],
        )

        # Submit a new WIP patch with reviewers.
        warnings, errors = submit.validate_commit_stack(
            [_commit(rev=None, request=["alice"], wip=True)], Args(wip=True)
        )
        self.assertEqual((warnings, errors), ({}, {}))

        # Submit a new patch with the WIP prefix.
        warnings, _ = submit.validate_commit_stack(
            [_commit(title="WIP: A", request=["alice"], wip=True)], Args()
        )
        self.assertEqual(
            list(warnings.values()), [[Contains('submitted as "Changes Planned"')]]
        )

        # Submitting a stack of two commits with one revision already on Phabricator
        # should properly set `commit.rev_phid`.
        commits = [_commit(title="submitted", rev=1), _commit(title="new", rev=None)]
        submit.validate_commit_stack(commits, Args())
        self.assertEqual(
            commits[0].rev_phid,
            "PHID-DREV-1",
            "`rev_id` should be set for already submitted commit.",
        )
        self.assertIsNone(
            commits[1].rev_phid, "`rev_id` should not be set for newly created commit."
        )

        # Submitting a new uplift clears reviewers and shouldn't warn.
        warnings, errors = submit.validate_commit_stack(
            [_commit(rev=None, granted=[], request=[])], Args(command="uplift")
        )
        self.assertEqual((warnings, errors), ({}, {}))

    def test_update_commits_from_args(self):
        def lwr(revs):
            return [r.lower() for r in revs]

        update = submit.update_commits_from_args

        _commits = [
            Commit(title="A", reviewers={"granted": [], "request": []}, bug_id=None),
            Commit(
                title="B",
                reviewers={"granted": [], "request": ["one"]},
                bug_id="1",
            ),
        ]

        # No change if noreviewer  args provided
        commits = copy.deepcopy(_commits)
        commits[1].reviewers["granted"].append("two")
        with mock.patch("mozphab.commands.submit.config") as m_config:
            m_config.always_blocking = False
            update(commits, Args())
            self.assertEqual(
                commits,
                [
                    Commit(
                        title="A",
                        reviewers={"granted": [], "request": []},
                        bug_id=None,
                        wip=True,
                    ),
                    Commit(
                        title="B",
                        reviewers={"granted": ["two"], "request": ["one"]},
                        bug_id="1",
                        wip=False,
                    ),
                ],
            )

        # Adding and removing reviewers, forcing the bug ID
        commits = copy.deepcopy(_commits)
        update(commits, Args(reviewer=["two", "three"], bug="2"))
        assert commits[0].title == "A"
        assert commits[0].bug_id == "2"
        assert "two" in commits[0].reviewers["granted"]
        assert "three" in commits[0].reviewers["granted"]

        assert commits[1].title == "B"
        assert commits[1].bug_id == "2"
        assert "two" in commits[1].reviewers["granted"]
        assert "three" in commits[1].reviewers["granted"]

        # Removing duplicates
        commits = copy.deepcopy(_commits)
        update(
            commits,
            Args(
                reviewer=["Two", "two", "two!", "three", "Three", "THREE!"],
                blocker=["Two", "THREE!", "three", "two", "three"],
            ),
        )
        assert commits[0].title == "A"
        assert commits[0].bug_id is None
        assert "two!" in lwr(commits[0].reviewers["granted"])
        assert "three!" in lwr(commits[0].reviewers["granted"])

        assert commits[1].title == "B"
        assert commits[1].bug_id == "1"
        assert "two!" in lwr(commits[1].reviewers["granted"])
        assert "three!" in lwr(commits[1].reviewers["granted"])

        # Adding blocking reviewers via args
        commits = copy.deepcopy(_commits)
        commits[1].reviewers["request"].append("three")
        commits[1].reviewers["granted"].append("four")
        commits[1].reviewers["granted"].append("five")
        update(
            commits,
            Args(
                reviewer=["one", "two!", "four"],
                blocker=["three", "four!"],
            ),
        )
        assert commits[0].title == "A"
        assert commits[0].bug_id is None
        assert "one" in lwr(commits[0].reviewers["granted"])
        assert "two!" in lwr(commits[0].reviewers["granted"])
        assert "three!" in lwr(commits[0].reviewers["granted"])
        assert "four!" in lwr(commits[0].reviewers["granted"])

        assert commits[1].title == "B"
        assert commits[1].bug_id == "1"
        assert "four!" in commits[1].reviewers["granted"]
        assert "two!" in commits[1].reviewers["granted"]
        assert "one" in commits[1].reviewers["request"]
        assert "three!" in commits[1].reviewers["request"]

        # reviewerless should result in WIP commits
        commits = copy.deepcopy(_commits)
        commits.append(Commit(title="Bug 2 - A", bug_id="2", rev_id=2))
        with mock.patch("mozphab.commands.submit.conduit.has_revision_reviewers") as m:
            m.side_effect = [
                False,  # First commit doesn't have reviewers.
                # Function isn't called for second commit since it has reviewers.
                True,  # For revision ID 2.
            ]
            update(commits, Args())
        assert commits[0].wip
        assert not commits[1].wip
        assert not commits[2].wip

        # Force WIP
        commits = copy.deepcopy(_commits)
        update(commits, Args(wip=True))
        assert commits[0].wip
        assert commits[1].wip

        # reviewerless with --no-wip shouldn't be WIP
        commits = copy.deepcopy(_commits)
        commits.append(Commit(title="WIP: Bug 2 - A", bug_id="2", wip=True))
        update(commits, Args(no_wip=True))
        assert not commits[0].wip
        assert not commits[1].wip
        assert not commits[2].wip

        # Reviewerless uplift commits should not be marked as WIP, since
        # uplift clears reviewers and `#release-managers` is set by automation.
        commits = copy.deepcopy(_commits)
        update(commits, Args(command="uplift"))
        assert not commits[0].wip, "Uplift commit without reviewers should not be WIP."
        assert not commits[1].wip, "Uplift commit with reviewers should not be WIP."

        # Forcing blocking reviewers
        commits = copy.deepcopy(_commits)
        commits[1].reviewers["granted"].append("two")
        with mock.patch("mozphab.commands.submit.config") as m_config:
            m_config.always_blocking = True
            update(commits, Args())
            self.assertEqual(
                commits,
                [
                    Commit(
                        title="A",
                        reviewers={"granted": [], "request": []},
                        bug_id=None,
                        wip=True,
                    ),
                    Commit(
                        title="B",
                        reviewers={"granted": ["two!"], "request": ["one!"]},
                        bug_id="1",
                        wip=False,
                    ),
                ],
            )

    def test_single_fails_with_end_rev(self):
        # --single is working with one SHA1 provided only
        repo = repository.Repository("", "", "dummy")

        self._assertNoError(repo.set_args, Args(single=True))
        self._assertError(
            repo.set_args,
            "Option --single can be used with only one identifier.",
            Args(single=True, end_rev="endrev"),
        )


class TestUpdateCommitSummary(unittest.TestCase):
    def test_update_revision_description(self):
        c = commit(
            rev_id="D123",
            title="hi!",
            body="hello!  µ-benchmarks are a thing.\n\n"
            "Differential Revision: http://phabricator.test/D123",
        )
        r = {"fields": {"title": "", "summary": ""}}
        t = []

        expected = [
            {"type": "title", "value": "hi!"},
            {"type": "summary", "value": "hello!  µ-benchmarks are a thing."},
        ]
        submit.update_revision_description(t, c, r)

        self.assertListEqual(t, expected)

    def test_update_revision_description_no_op(self):
        c = commit(
            rev_id="D123",
            title="hi!",
            body="hello!\n\nDifferential Revision: http://phabricator.test/D123",
        )
        r = {"fields": {"title": "hi!", "summary": "hello!\n\n"}}
        t = []

        expected = []
        submit.update_revision_description(t, c, r)

        self.assertListEqual(t, expected)

    @mock.patch("mozphab.conduit.ConduitAPI.get_users")
    def test_update_revision_reviewers(self, m_get_users):
        # From https://phabricator.services.mozilla.com/api/differential.revision.edit
        #
        # Example call format we are aiming for:
        #
        # $ echo '{
        #   "transactions": [
        #     {
        #       "type": "reviewers.set",
        #       "value": ["PHID-USER-1", "blocking(PHID-USER-2)"]
        #     }
        #   ],
        #   "objectIdentifier": "D8095"
        # }' | arc call-conduit --conduit-uri \
        #       https://phabricator.services.mozilla.com/ \
        #       --conduit-token <conduit-token> differential.revision.edit
        m_get_users.side_effect = (
            [
                {"phid": "PHID-USER-1"},
                {"phid": "PHID-USER-2"},
                {"phid": "PHID-USER-3"},
            ],
            [{"phid": "PHID-USER-1"}],
            [{"phid": "PHID-USER-2"}, {"phid": "PHID-USER-3"}],
        )
        c = commit(rev_id="123", reviewers=[["alice", "bob!"], ["frankie!"]])
        t = []

        expected = [
            {
                "type": "reviewers.set",
                "value": [
                    "PHID-USER-1",
                    "blocking(PHID-USER-2)",
                    "blocking(PHID-USER-3)",
                ],
            }
        ]
        mozphab.conduit.update_revision_reviewers(t, c)

        self.assertEqual(
            m_get_users.call_args_list,
            [
                mock.call(["alice", "bob!", "frankie!"]),
                mock.call(["alice"]),
                mock.call(["bob", "frankie"]),
            ],
        )
        self.assertListEqual(t, expected)

    def test_parse_api_response_with_no_problems(self):
        # Response comes from running:
        # $ echo '{... (some valid update summary JSON) ...}' | \
        #   arc call-conduit differential.revision.edit
        api_response = (
            '{"error":null,"errorMessage":null,"response":{"object":{'
            '"id":56,"phid":"PHID-DREV-ke6jhbdnwd5chtnk2q5w"},'
            '"transactions":[{"phid":"PHID-XACT-DREV-itlxgx7rsjrcnta"}]}} '
        )
        self.assertEqual(None, helpers.parse_api_error(api_response))

    def test_parse_api_response_with_errors(self):
        # Error response from running:
        # $ echo '{}' | arc call-conduit differential.revision.edit
        api_response = (
            '{"error":"ERR-CONDUIT-CORE", '
            '"errorMessage":"ERR-CONDUIT-CORE: Parameter '
            '\\"transactions\\" is not a list of transactions.",'
            '"response":null} '
        )
        self.assertEqual(
            'ERR-CONDUIT-CORE: Parameter "transactions" is not a list of transactions.',
            helpers.parse_api_error(api_response),
        )

    def test_update_revision_no_bug_id(self):
        # Phabricator stores patches with no bug as having an empty string as the
        # bug ID. We should not explicitly update the bug-id if our bug id is "None".
        transactions = []
        submit.update_revision_bug_id(
            transactions, Commit(bug_id=None), {"fields": {"bugzilla.bug-id": ""}}
        )
        self.assertEqual([], transactions)


if __name__ == "__main__":
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse

from mozphab.conduit import ConduitAPIError, conduit
from mozphab.environment import INSTALL_CERT_MSG
from mozphab.helpers import get_arcrc_path, read_json_field
from mozphab.logger import logger
from mozphab.repository import Repository


def doctor(repo: Repository, args: argparse.Namespace):
    """Validates the user's installation of moz-phab.

    Currently, this only retrieves the user's Phabricator
    API key and validates it against the API.

    Raises ConduitAPIError if token is invalid in any way.
    """
    token = read_json_field([get_arcrc_path()], ["hosts", repo.api_url, "token"])
    if not token:
        raise ConduitAPIError(INSTALL_CERT_MSG)

    try:
        who = conduit.whoami(api_token=token)
        logger.info(f"Phabricator API token for {who['userName']} is valid.")
    except ConduitAPIError as e:
        if "not valid" in str(e):
            logger.error(e)
            raise ConduitAPIError(INSTALL_CERT_MSG) from e

        # Re-raise the exception if it is N/A to the above conditions.
        raise e

    repo.validate_email()

    logger.info("No critical issues detected with moz-phab installation.")


def add_parser(parser):
    doctor_parser = parser.add_parser(
        "doctor", help="Ensure your moz-phab installation is valid."
    )
    doctor_parser.add_argument(
        "--safe-mode",
        dest="safe_mode",
        action="store_true",
        help="Run VCS with only necessary extensions.",
    )
    doctor_parser.set_defaults(func=doctor, needs_repo=True)



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import base64
import concurrent.futures
import datetime
import hashlib
import json
import operator
import os
import urllib.parse as url_parse
import urllib.request as url_request
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from .commits import Commit
from .diff import Diff
from .environment import INSTALL_CERT_MSG, USER_AGENT
from .exceptions import (
    CommandError,
    Error,
    NonLinearException,
    NotFoundError,
)
from .helpers import (
    get_arcrc_path,
    read_json_field,
    strip_differential_revision,
)
from .logger import logger
from .simplecache import cache


def normalise_reviewer(reviewer: str, strip_group: bool = True) -> str:
    """This provide a canonical form of the reviewer for comparison."""
    reviewer = reviewer.rstrip("!").lower()
    if strip_group:
        reviewer = reviewer.lstrip("#")
    return reviewer


class ConduitAPIError(Error):
    """Raised when the Phabricator Conduit API returns an error response."""

    def __init__(self, msg: Optional[str] = None):
        super().__init__(f"Phabricator Error: {msg if msg else 'Unknown Error'}")


class ConduitAPI:
    def __init__(self):
        self.repo = None

    def set_repo(self, repo):
        self.repo = repo

    @property
    def repo_phid(self) -> str:
        return self.repo.phid

    def load_api_token(self) -> str:
        """Return an API Token for the given repository.

        Returns:
            API Token string
        """
        try:
            return os.environ["MOZPHAB_PHABRICATOR_API_TOKEN"]
        except KeyError:
            pass

        if "api_token" in cache:
            return str(cache.get("api_token"))

        token = read_json_field(
            [get_arcrc_path()], ["hosts", self.repo.api_url, "token"]
        )
        if not token:
            raise ConduitAPIError(INSTALL_CERT_MSG)
        cache.set("api_token", token)
        return token

    def save_api_token(self, token: str):
        filename = get_arcrc_path()
        created = False
        try:
            with open(filename, "r", encoding="utf-8") as f:
                rc = json.load(f)
        except FileNotFoundError:
            rc = {}
            created = True

        rc.setdefault("hosts", {})
        rc["hosts"].setdefault(self.repo.api_url, {})
        rc["hosts"][self.repo.api_url]["token"] = token

        with open(filename, "w", encoding="utf-8") as f:
            json.dump(rc, f, sort_keys=True, indent=2)

        if created:
            os.chmod(filename, 0o600)

    def call(
        self, api_method: str, api_call_args: dict, *, api_token: Optional[str] = None
    ) -> dict:
        """Call Conduit API and return the JSON API call result.

        Args:
            api_method: The API method name to call, like 'differential.revision.edit'.
            api_call_args: JSON dict of call args to send.
            api_token: Use specific token to authentication instead of saved value.

        Returns:
            JSON API call result object

        Raises:
            ConduitAPIError if the API threw an error back at us.
        """
        req_args = self._build_request(
            method=api_method,
            args=api_call_args,
            token=api_token,
        )
        logger.debug("%s %s", req_args["url"], api_call_args)

        with url_request.urlopen(url_request.Request(**req_args)) as r:
            res = json.load(r)
        if res["error_code"]:
            raise ConduitAPIError(res.get("error_info", "Error %s" % res["error_code"]))
        return res["result"]

    def _build_request(
        self, *, method: str, args: dict, token: Optional[str]
    ) -> Dict[str, Any]:
        """Return dict with Request args for calling the specified conduit method."""
        return {
            "url": url_parse.urljoin(self.repo.api_url, method),
            "method": "POST",
            "headers": {"User-Agent": USER_AGENT},
            "data": url_parse.urlencode(
                {
                    "params": json.dumps(
                        {
                            **args,
                            "__conduit__": {"token": token or self.load_api_token()},
                        },
                        separators=(",", ":"),
                    ),
                    "output": "json",
                    "__conduit__": True,
                }
            ).encode(),
        }

    def ping(self) -> bool:
        """Sends a ping to the Phabricator server using `conduit.ping` API.

        Returns: `True` if no error, otherwise - `False`
        """
        try:
            self.call("conduit.ping", {})
        except ConduitAPIError as err:
            logger.error(err)
            return False
        except CommandError as err:
            logger.error(err)
            return False
        return True

    def check(self) -> bool:
        """Check if raw Conduit API can be used."""
        # Check if the cache file exists
        path = os.path.join(self.repo.dot_path, ".moz-phab_conduit-configured")
        if os.path.isfile(path):
            return True

        if self.ping():
            # Create the cache file
            with open(path, "a"):
                os.utime(path, None)
            return True

        return False

    def get_projects(self, slugs: List[str]) -> List[dict]:
        """Search for tags by hashtags."""
        response = self.call("project.search", {"constraints": {"slugs": slugs}})
        return response.get("data")

    def get_project_phid(self, slug: str) -> Optional[str]:
        projects = self.get_projects([slug])
        if not projects:
            return None

        return projects[0]["phid"]

    def ids_to_phids(self, rev_ids: List[int]) -> List[str]:
        """Convert revision ids to PHIDs.

        Parameters:
            rev_ids (list): A list of revision ids

        Returns:
            A list of PHIDs.
        """
        return [r["phid"] for r in self.get_revisions(ids=rev_ids)]

    def id_to_phid(self, rev_id: int) -> str:
        """Convert revision id to PHID."""
        phids = self.ids_to_phids([rev_id])
        if phids:
            return phids[0]

        raise NotFoundError("revision {} not found".format(rev_id))

    def phids_to_ids(self, phids: List[str]) -> List[str]:
        """Convert revision PHIDs to ids.

        Parameteres:
            phids (list): A list of PHIDs

        Returns:
            A list of ids.
        """
        return ["D{}".format(r["id"]) for r in self.get_revisions(phids=phids)]

    def phid_to_id(self, phid: str) -> str:
        """Convert revision PHID to id."""
        ids = self.phids_to_ids([phid])
        if ids:
            return ids[0]

        raise NotFoundError("revision {} not found".format(phid))

    def get_revisions(
        self, ids: Optional[List[int]] = None, phids: Optional[List[int]] = None
    ) -> List[dict]:
        """Get revisions info from Phabricator.

        Args:
            ids - list of revision ids
            phids - list of revision phids

        Returns a list of revisions ordered by ids or phids
        """
        if (ids and phids) or (ids is None and phids is None):
            raise ValueError("Internal Error: Invalid args to get_revisions")

        # Initialise depending on if we're passed revision IDs or PHIDs.
        if ids:
            ids = [str(rev_id) for rev_id in ids]
            phids_by_id = {
                rev_id: cache.get("rev-id-%s" % rev_id)
                for rev_id in ids
                if "rev-id-%s" % rev_id in cache
            }
            found_phids = list(phids_by_id.values())
            query_field = "ids"
            query_values = [
                int(rev_id) for rev_id in set(ids) - set(phids_by_id.keys())
            ]

        else:
            phids_by_id = {}
            found_phids = phids.copy()
            query_field = "phids"
            query_values = {phid for phid in phids if "rev-%s" % phid not in cache}

        # Revisions metadata keyed by PHID.
        revisions = {
            phid: cache.get("rev-%s" % phid)
            for phid in found_phids
            if "rev-%s" % phid in cache
        }

        # Query Phabricator if we don't have cached values for revisions.
        if query_values:
            api_call_args = {
                "constraints": {query_field: sorted(query_values)},
                "attachments": {"reviewers": True},
            }
            response = self.call("differential.revision.search", api_call_args)
            rev_list = response.get("data")

            for r in rev_list:
                phids_by_id[str(r["id"])] = r["phid"]
                revisions[r["phid"]] = r
                cache.set("rev-id-%s" % r["id"], r["phid"])
                cache.set("rev-%s" % r["phid"], r)

        # Return revisions in the same order requested.
        if ids:
            # Skip revisions for which we do not have a query result.
            return [
                revisions[phids_by_id[rev_id]]
                for rev_id in ids
                if rev_id in phids_by_id
            ]
        else:
            return [revisions[phid] for phid in phids]

    def get_diffs(
        self, ids: Optional[List[int]] = None, phids: Optional[List[str]] = None
    ) -> Dict[str, Dict]:
        """Get diffs from Phabricator.

        Args:
            ids - a list of diff IDs to pull
            phids - a list of diff PHIDs to pull

        Returns a dict of diffs identified by their PHID
        """
        if (ids and phids) or (ids is None and phids is None):
            raise ValueError("Internal Error: Invalid args to get_diffs")

        if ids:
            constraints = {"ids": list(set(ids))}
        else:
            constraints = {"phids": list(set(phids))}

        api_call_args = {
            "constraints": constraints,
            "attachments": {"commits": True},
        }
        response = self.call("differential.diff.search", api_call_args)
        diff_list = response.get("data", [])

        diff_dict = {}
        for diff in diff_list:
            diff_dict[diff["phid"]] = diff

        return diff_dict

    def get_successor_phids(
        self, phid: str, include_abandoned: bool = False
    ) -> List[str]:
        return self.get_related_phids(
            phid, relation="child", include_abandoned=include_abandoned
        )

    def get_ancestor_phids(
        self, phid: str, include_abandoned: bool = False
    ) -> List[str]:
        return self.get_related_phids(
            phid, relation="parent", include_abandoned=include_abandoned
        )

    def get_related_phids(
        self, base_phid: str, relation: str = "parent", include_abandoned: bool = False
    ) -> List[str]:
        """Returns the list of PHIDs with direct dependency"""
        result = []

        def _get_related(phid):
            api_call_args = {"sourcePHIDs": [phid], "types": ["revision.%s" % relation]}
            edge = self.call("edge.search", api_call_args)
            if edge.get("data"):
                if len(edge["data"]) > 1:
                    raise NonLinearException()

                result.append(edge["data"][0]["destinationPHID"])
                _get_related(result[-1])

        _get_related(base_phid)

        if not result or include_abandoned:
            return result

        revisions = self.get_revisions(phids=result)
        return [
            r["phid"]
            for r in revisions
            if r["fields"]["status"]["value"] != "abandoned"
        ]

    def get_users(self, usernames: List[str]) -> List[dict]:
        """Get users using the user.query API.

        Caches the result in the process.
        Returns a list of existing Phabricator users data.
        """
        to_collect = []
        users = []
        for user in usernames:
            u = user.rstrip("!")
            key = "user-%s" % u
            if key in cache:
                users.append(cache.get(key))
            else:
                to_collect.append(u)

        if not to_collect:
            return users

        api_call_args = {"usernames": to_collect}
        # We're using the deprecated user.query API as the user.search does not
        # provide the user availability information.
        # See https://phabricator.services.mozilla.com/conduit/method/user.query/
        response = self.call("user.query", api_call_args)
        for user in response:
            users.append(user)
            key = "user-%s" % user["userName"]
            cache.set(key, user)
            cache.set(user["phid"], key)

        return users

    def get_groups(self, slugs: List[str]) -> List[dict]:
        to_collect = []
        groups = []
        for slug in slugs:
            s = slug.rstrip("!")
            key = "group-%s" % s
            if key in cache:
                groups.append(cache.get(key))
            else:
                to_collect.append(s)

        if not to_collect:
            return groups

        # See https://phabricator.services.mozilla.com/conduit/method/project.search/
        api_call_args = {"queryKey": "active", "constraints": {"slugs": to_collect}}
        response = self.call("project.search", api_call_args)
        for data in response.get("data"):
            group = {"name": data["fields"]["slug"], "phid": data["phid"]}
            groups.append(group)
            key = "group-%s" % group["name"]
            cache.set(key, group)

        # projects might be received by an alias.
        maps = response["maps"]["slugMap"]
        for alias in maps.keys():
            name = normalise_reviewer(alias)
            group = {"name": name, "phid": maps[alias]["projectPHID"]}
            key = "group-%s" % alias
            if key not in cache:
                groups.append(group)
                cache.set(key, group)

        return groups

    def create_revision(
        self,
        commit: Commit,
        diff_phid: str,
        parent_rev_phid: Optional[str] = None,
    ) -> dict:
        """Create a new revision in Phabricator."""
        transactions = [
            {"type": "title", "value": commit.revision_title()},
            {"type": "summary", "value": commit.body},
        ]
        if commit.has_reviewers and not commit.wip:
            self.update_revision_reviewers(transactions, commit)

        if commit.bug_id:
            transactions.append({"type": "bugzilla.bug-id", "value": commit.bug_id})

        return self.edit_revision(
            transactions=transactions,
            diff_phid=diff_phid,
            wip=commit.wip,
            parent_rev_phid=parent_rev_phid,
        )

    def update_revision(
        self,
        commit: Commit,
        diff_phid: Optional[str] = None,
        comment: Optional[str] = None,
        parent_rev_phid: Optional[str] = None,
    ) -> dict:
        """Update an existing revision in Phabricator."""
        # Update the title and summary
        transactions = [
            {"type": "title", "value": commit.revision_title()},
            {"type": "summary", "value": strip_differential_revision(commit.body)},
        ]

        # Add update comment
        if comment:
            transactions.append({"type": "comment", "value": comment})

        # Add reviewers only if revision lacks them
        if (
            commit.has_reviewers
            and not commit.wip
            and not conduit.has_revision_reviewers(commit)
        ):
            self.update_revision_reviewers(transactions, commit)

        # Update bug id if different
        if commit.bug_id:
            revision = conduit.get_revisions(ids=[commit.rev_id])[0]
            if revision["fields"]["bugzilla.bug-id"] != commit.bug_id:
                transactions.append({"type": "bugzilla.bug-id", "value": commit.bug_id})

        return self.edit_revision(
            transactions=transactions,
            diff_phid=diff_phid,
            rev_id=commit.rev_id,
            wip=commit.wip,
            parent_rev_phid=parent_rev_phid,
        )

    def edit_revision(
        self,
        transactions: Optional[List[dict]] = None,
        diff_phid: Optional[str] = None,
        rev_id: Optional[str] = None,
        wip: bool = False,
        parent_rev_phid: Optional[str] = None,
    ) -> dict:
        """Edit (create or update) a revision."""
        trans = transactions or []
        post_trans = []

        # diff_phid is not present for changes in revision settings (like WIP)
        if diff_phid:
            trans.append({"type": "update", "value": diff_phid})

        # Add revision parents.
        if parent_rev_phid:
            trans.append({"type": "parents.add", "value": [parent_rev_phid]})

        existing_status = None
        if rev_id:
            try:
                args = {"ids": [int(rev_id)]}
            except ValueError:
                args = {"phids": [rev_id]}
            existing_revision = conduit.get_revisions(**args)[0]
            existing_status = existing_revision["fields"]["status"]["value"]

        # Set revision for changes-planned or needs-review as required.
        # Phabricator will throw an error if we attempt to set a status to the same
        # as the current status.
        if wip:
            # Phabricator will automatically set the revision to needs-review
            # after the call to differential.revision.edit. If we are a creating a new
            # revision, we can add the `plan-changes` transaction to our create call.
            # Existing revisions must make a subsequent API call to set the status to
            # `changes-planned` to match our WIP state.
            plan_changes_transaction = {"type": "plan-changes", "value": True}
            if existing_status == "changes-planned":
                post_trans.append(plan_changes_transaction)
            else:
                trans.append(plan_changes_transaction)

        # Let the Phabricator sort out the correct status for new revisions to avoid
        # emails being sent before phabbugs has processed the revision.
        # Updates to existing reviews that have been accepted shouldn't be triggered
        # for re-review.
        elif existing_status and existing_status not in ("needs-review", "accepted"):
            trans.append({"type": "request-review", "value": True})

        # Call differential.revision.edit
        api_call_args = {"transactions": trans}
        if rev_id:
            api_call_args["objectIdentifier"] = rev_id
        revision = self.call("differential.revision.edit", api_call_args)
        if not revision:
            raise ConduitAPIError("Can't edit the revision.")

        # Run post-edit transactions if required.
        if post_trans:
            revision = self.call(
                "differential.revision.edit",
                {
                    "objectIdentifier": rev_id,
                    "transactions": post_trans,
                },
            )

        return revision

    def apply_transactions_to_revision(self, rev_id: str, transactions: List[dict]):
        """Apply transactions to the specified revision."""
        self.call(
            "differential.revision.edit",
            {"objectIdentifier": rev_id, "transactions": transactions},
        )

    def get_repository_by_callsign(self, call_sign: str) -> dict:
        """Get repository info for a repo on Phabricator by callsign."""
        return self.repository_search_single("callsigns", call_sign)

    def get_repository_by_shortname(self, short_name: str) -> dict:
        """Get repository info for a repo on Phabricator by shortname."""
        return self.repository_search_single("shortNames", short_name)

    def repository_search_single(self, constraint: str, value: str) -> dict:
        """Get the information about a single repository from Phabricator.

        Takes a `constraint` to be passed to `diffusion.repository.search`
        and a `value` to be passed as the value of `constraint`. `value`
        must be a single element and will be passed in a list to Conduit.
        """
        api_call_args = {"constraints": {constraint: [value]}, "limit": 1}
        data = self.call("diffusion.repository.search", api_call_args)
        if not data.get("data"):
            raise NotFoundError("Repository %s not found" % value)

        repo = data["data"][0]
        return repo

    def get_repositories_with_tag(self, tag: str) -> dict:
        """Get repository information for repos associated with the given tag."""
        api_call_args = {
            "constraints": {
                "projects": [tag],
            }
        }

        data = self.call("diffusion.repository.search", api_call_args).get("data")
        if not data:
            raise NotFoundError(f"No repositories found with tag {tag}")

        return data

    def build_creation_method(self) -> str:
        """Build the creationMethod string for differential.creatediff."""
        creation_method = ["moz-phab", self.repo.vcs]

        if (
            self.repo.vcs == "git" or self.repo.vcs == "jj"
        ) and self.repo.is_cinnabar_required:
            creation_method.append("cinnabar")

        # Add uplift indicator if this is an uplift command
        if self.repo.args.command == "uplift":
            creation_method.append("uplift")

        # Indicate Lando as the creation method when the
        # API key is set programmatically from the environment.
        # In theory a user could set this variable themselves,
        # so we should come up with a better way to check this
        # in the future.
        if "MOZPHAB_PHABRICATOR_API_TOKEN" in os.environ:
            creation_method.append("lando")

        return "-".join(creation_method)

    def submit_diff(self, diff: Diff, commit: Commit) -> dict:
        files_changed = sorted(
            diff.changes.values(), key=operator.attrgetter("cur_path")
        )
        changes = [
            change.to_conduit(conduit.repo.get_public_node(commit.node))
            for change in files_changed
        ]

        base_revision = conduit.repo.get_public_node(commit.parent)

        creation_method_str = self.build_creation_method()

        # Use the repo of the existing revision if this is an update.
        revs = conduit.get_revisions(ids=[commit.rev_id]) if commit.rev_id else []
        diff_repo_phid = revs[0]["fields"]["repositoryPHID"] if revs else self.repo_phid

        api_call_args = {
            "changes": changes,
            "sourceMachine": self.repo.phab_url,
            "sourceControlSystem": self.repo.phab_vcs,
            "sourceControlPath": "/",
            "sourceControlBaseRevision": base_revision,
            "creationMethod": creation_method_str,
            "lintStatus": "none",
            "unitStatus": "none",
            "repositoryPHID": diff_repo_phid,
            "sourcePath": self.repo.path,
            "branch": "HEAD" if self.repo.phab_vcs == "git" else "default",
        }
        return self.call("differential.creatediff", api_call_args)

    def set_diff_property(self, diff_id: str, commit: Commit, message: str):
        """Add information about our local commit to the diff."""
        data = {
            commit.node: {
                "author": commit.author_name,
                "authorEmail": commit.author_email,
                "time": commit.author_date_epoch,
                "summary": commit.revision_title(),
                "message": message,
                "commit": conduit.repo.get_public_node(commit.node),
                "parents": [conduit.repo.get_public_node(commit.parent)],
            }
        }
        if commit.tree_hash is not None:
            data[commit.node]["tree"] = commit.tree_hash

        if self.repo.phab_vcs == "hg":
            data[commit.node]["rev"] = commit.node

        api_call_args = {
            "diff_id": diff_id,
            "name": "local:commits",
            "data": json.dumps(data),
        }
        self.call("differential.setdiffproperty", api_call_args)

    def upload_files_from_diff(self, diff: Diff):
        # Files are uploaded in parallel using a pool of threads.
        with concurrent.futures.ThreadPoolExecutor(max_workers=5) as executor:
            futures = []
            for change in diff.changes.values():
                for upload in change.uploads:
                    path = (
                        change.cur_path if upload["type"] == "new" else change.old_path
                    )
                    futures.append(executor.submit(self.upload_file, upload, path))

            # Wait for all uploads to be finished.
            concurrent.futures.wait(futures)

            # Check that all went well. If not, propagate the first error here
            # by calling the future's result() method.
            for upload in futures:
                upload.result()

    def upload_file(self, upload: dict, path: str):
        data: bytes = upload["value"]
        if not data:
            return
        name = os.path.basename(path)
        allocation = self.call(
            "file.allocate",
            {
                "name": name,
                "contentLength": len(data),
                "contentHash": hashlib.sha256(data).hexdigest(),
            },
        )
        file_phid = allocation["filePHID"]
        if allocation["upload"]:
            if not file_phid:
                data_base64 = base64.standard_b64encode(data)
                file_phid = self.call(
                    "file.upload", {"data_base64": data_base64.decode(), "name": name}
                )
            else:
                chunks = self.call("file.querychunks", {"filePHID": file_phid})
                for chunk in chunks:
                    if chunk["complete"]:
                        continue
                    byte_start = int(chunk["byteStart"])
                    byte_end = int(chunk["byteEnd"])
                    data_base64 = base64.standard_b64encode(data[byte_start:byte_end])
                    self.call(
                        "file.uploadchunk",
                        {
                            "filePHID": file_phid,
                            "byteStart": byte_start,
                            "data": data_base64.decode(),
                            "dataEncoding": "base64",
                        },
                    )

        upload["phid"] = str(file_phid)

    def whoami(self, *, api_token: Optional[str] = None) -> dict:
        if "whoami" in cache:
            return dict(cache.get("whoami"))

        who = self.call("user.whoami", {}, api_token=api_token)
        cache.set("whoami", who)
        return who

    def has_revision_reviewers(self, commit: Commit) -> bool:
        """Return True if the commit has a remote revision with reviewers.

        If the commit doesn't have a revision ID or the revision cannot be found, then
        return False.
        """
        if not commit.rev_id:
            return False
        revs = self.get_revisions(ids=[commit.rev_id])
        return bool(revs and revs[0]["attachments"]["reviewers"]["reviewers"])

    def update_revision_reviewers(
        self, transactions: List[Dict[str, Any]], commit: Commit
    ):
        # Appends differential.revision.edit transaction(s) to `transactions` to
        # set the reviewers.

        all_reviewing = commit.reviewers["request"] + commit.reviewers["granted"]

        # Find reviewers PHIDs
        all_reviewers = [r for r in all_reviewing if not r.startswith("#")]
        # preload all reviewers
        self.get_users(all_reviewers)
        reviewers = [r for r in all_reviewers if not r.endswith("!")]
        blocking_reviewers = [r.rstrip("!") for r in all_reviewers if r.endswith("!")]
        reviewers_phid = [user["phid"] for user in self.get_users(reviewers)]
        blocking_phid = [
            "blocking(%s)" % user["phid"] for user in self.get_users(blocking_reviewers)
        ]

        # Find groups PHIDs
        all_groups = [g for g in all_reviewing if g.startswith("#")]
        groups = [g for g in all_groups if not g.endswith("!")]
        blocking_groups = [g.rstrip("!") for g in all_groups if g.endswith("!")]
        # preload all groups
        self.get_groups(all_groups)
        groups_phid = [group["phid"] for group in self.get_groups(groups)]
        bl_groups_phid = [
            "blocking(%s)" % group["phid"] for group in self.get_groups(blocking_groups)
        ]

        all_reviewing_phid = (
            reviewers_phid + blocking_phid + groups_phid + bl_groups_phid
        )
        transactions.extend([{"type": "reviewers.set", "value": all_reviewing_phid}])

    def check_for_invalid_reviewers(self, reviewers: dict) -> List[Dict[str, Any]]:
        """Return a list of invalid reviewer names.

        Args:
            reviewers: A commit reviewers dict of granted and requested reviewers.
        """

        # Combine the lists of requested reviewers and granted reviewers.
        all_reviewers = []
        found_names = []
        for sublist in list(reviewers.values()):
            all_reviewers.extend(
                [
                    normalise_reviewer(r, strip_group=False)
                    for r in sublist
                    if not r.startswith("#")
                ]
            )

        users = []
        if all_reviewers:
            users = self.get_users(all_reviewers)
            found_names = [
                normalise_reviewer(data["userName"], strip_group=False)
                for data in users
            ]

        # Group reviewers are represented by a "#" prefix
        all_groups = []
        found_groups = []
        for sublist in list(reviewers.values()):
            all_groups.extend(
                [
                    normalise_reviewer(r, strip_group=False)
                    for r in sublist
                    if r.startswith("#")
                ]
            )

        if all_groups:
            groups = self.get_groups(all_groups)
            found_groups = [
                "#%s" % normalise_reviewer(group["name"]) for group in groups
            ]

        all_reviewers.extend(all_groups)
        found_names.extend(found_groups)
        invalid = list(set(all_reviewers) - set(found_names))

        # Find users availability:
        unavailable = [
            {
                "name": r["userName"],
                "until": datetime.datetime.fromtimestamp(
                    r["currentStatusUntil"]
                ).strftime("%Y-%m-%d %H:%M"),
            }
            for r in users
            if r.get("currentStatus") == "away"
        ]

        # Find disabled users:
        disabled = [
            {"name": r["userName"], "disabled": True}
            for r in users
            if "disabled" in r.get("roles", [])
        ]
        return disabled + unavailable + [{"name": r} for r in invalid]


conduit = ConduitAPI()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import unittest

from mozphab import helpers
from mozphab.commands import submit
from mozphab.commits import Commit


class CommitParsing(unittest.TestCase):
    def assertParsed(self, result, parsed):
        self.assertEqual({"request": result[0], "granted": result[1]}, parsed)

    def test_bug_id(self):
        parse = helpers.parse_bugs

        self.assertEqual(parse("bug 1"), ["1"])
        self.assertEqual(parse("bug 123456"), ["123456"])
        self.assertEqual(parse("testb=1234x"), [])
        self.assertEqual(parse("ab4665521e2f"), [])
        self.assertEqual(parse("Aug 2008"), [])
        self.assertEqual(parse("GECKO_191a2_20080815_RELBRANCH"), [])
        self.assertEqual(parse("12345 is a bug"), [])
        self.assertEqual(parse(" bug   123456 whitespace!"), ["123456"])
        self.assertEqual(parse("bug 1 and bug 2"), ["1", "2"])
        self.assertEqual(parse("BUG 1 helper_bug2.html"), ["1", "2"])

    def test_reviewers(self):
        parse = helpers.parse_reviewers

        # first with r? reviewer request syntax
        self.assertParsed((["romulus"], []), parse("stuff; r?romulus"))
        self.assertParsed((["#romulus"], []), parse("stuff; r?#romulus"))
        self.assertParsed((["rom"], []), parse("stuff; r?rom#ulus"))
        self.assertParsed(
            (["romulus", "remus"], []), parse("stuff; r?romulus, r?remus")
        )
        self.assertParsed((["romulus", "remus"], []), parse("stuff; r?romulus,r?remus"))
        self.assertParsed((["romulus", "remus"], []), parse("stuff; r?romulus, remus"))
        self.assertParsed((["romulus", "remus"], []), parse("stuff; r?romulus,remus"))
        self.assertParsed((["romulus"], []), parse("stuff; (r?romulus)"))
        self.assertParsed((["romulus", "remus"], []), parse("stuff; (r?romulus,remus)"))
        self.assertParsed((["romulus"], []), parse("stuff; [r?romulus]"))
        self.assertParsed(
            (["remus", "romulus"], []), parse(" stuff; [r?remus, r?romulus]")
        )
        self.assertParsed((["romulus"], []), parse("stuff; r?romulus, a=test-only"))
        self.assertParsed((["romulus"], []), parse("stuff; r?romulus, ux-r=test-only"))

        # now with r= mozphab granted syntax
        self.assertParsed(([], ["romulus"]), parse("stuff; r=romulus"))
        self.assertParsed(([], ["#romulus"]), parse("stuff; r=#romulus"))
        self.assertParsed(([], ["rom"]), parse("stuff; r=rom#ulus"))
        self.assertParsed(
            ([], ["romulus", "remus"]), parse("stuff; r=romulus, r=remus")
        )
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff; r=romulus,r=remus"))
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff; r=romulus, remus"))
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff; r=romulus,remus"))
        self.assertParsed(([], ["romulus"]), parse("stuff; (r=romulus)"))
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff; (r=romulus,remus)"))
        self.assertParsed(([], ["romulus"]), parse("stuff; [r=romulus]"))
        self.assertParsed(
            ([], ["remus", "romulus"]), parse("stuff; [r=remus, r=romulus]")
        )
        self.assertParsed(([], ["romulus"]), parse("stuff; r=romulus, a=test-only"))
        self.assertParsed(([], ["romulus"]), parse("stuff; r=romulus, a?test-only"))
        self.assertParsed(([], ["romulus"]), parse("stuff; r=romulus, ux-r=test-only"))

        # mixed r? and r=
        self.assertParsed((["romulus"], ["remus"]), parse("stuff; r?romulus r=remus"))
        self.assertParsed((["remus"], ["romulus"]), parse("stuff; r=romulus r?remus"))
        self.assertParsed(
            (["romulus", "gps"], ["remus"]), parse("stuff; r?romulus,gps r=remus")
        )
        self.assertParsed(
            (["romulus"], ["remus", "gps"]), parse("stuff; r?romulus r=remus,gps")
        )
        self.assertParsed(
            (["romulus", "next"], ["remus", "gps"]),
            parse("stuff; r?romulus r=remus r?next r=gps"),
        )
        self.assertParsed(
            (["romulus", "romulus!", "romulus"], ["romulus", "romulus!", "romulus"]),
            parse("stuff; r?romulus r=romulus r?romulus! r=romulus!,romulus r?romulus"),
        )
        self.assertParsed(
            (["remus", "gps"], ["romulus"]), parse("stuff; r=romulus r?remus,gps")
        )
        self.assertParsed(
            (["remus"], ["romulus", "gps"]), parse("stuff; r=romulus,gps r?remus")
        )

        # try some other separators than ;
        self.assertParsed(([], ["romulus"]), parse("stuff r=romulus"))
        self.assertParsed(
            ([], ["romulus", "remus"]), parse("stuff. r=romulus, r=remus")
        )
        self.assertParsed(
            ([], ["romulus", "remus"]), parse("stuff - r=romulus,r=remus")
        )
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff, r=romulus, remus"))
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff.. r=romulus,remus"))
        self.assertParsed(([], ["romulus"]), parse("stuff | (r=romulus)"))

        # make sure things work with different spacing
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff;r=romulus,r=remus"))
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff.r=romulus, r=remus"))
        self.assertParsed(([], ["romulus", "remus"]), parse("stuff,r=romulus, remus"))
        self.assertParsed(([], ["gps"]), parse("stuff; r=gps DONTBUILD (NPOTB)"))
        self.assertParsed(([], ["gps"]), parse("stuff; r=gps DONTBUILD"))
        self.assertParsed(([], ["gps"]), parse("stuff; r=gps (DONTBUILD)"))

        # make sure things work with a period in the nickname
        self.assertParsed(
            ([], ["jimmy.james", "bill.mcneal"]),
            parse("stuff;r=jimmy.james,r=bill.mcneal"),
        )

        # make sure period at the end does not get included
        self.assertParsed(([], ["jimmy"]), parse("stuff;r=jimmy."))

        # check some funky names too
        self.assertParsed(([], ["a"]), parse("stuff;r=a"))
        self.assertParsed(([], ["aa"]), parse("stuff;r=aa"))
        self.assertParsed(([], [".a"]), parse("stuff;r=.a"))
        self.assertParsed(([], ["..a"]), parse("stuff;r=..a"))
        self.assertParsed(([], ["...a"]), parse("stuff;r=...a"))
        self.assertParsed(([], ["a...a"]), parse("stuff;r=a...a"))
        self.assertParsed(([], ["a.b"]), parse("stuff;r=a.b"))
        self.assertParsed(([], ["a.b.c"]), parse("stuff;r=a.b.c"))
        self.assertParsed(([], ["-.-.-"]), parse("stuff;r=-.-.-"))

        # altogether now with some spaces sprinkled here and there
        self.assertParsed(
            (
                [],
                [
                    "a",
                    "aa",
                    ".a",
                    "..a",
                    "...a",
                    "a...a",
                    "a.b",
                    "a.b.c",
                    "-.-.-",
                ],
            ),
            parse("stuff;r=a,aa,.a,..a,...a, a...a,a.b, a.b.c, -.-.-"),
        )

        # bare r?
        self.assertParsed(([], []), parse("stuff; r?"))
        self.assertParsed(([], []), parse("stuff, r="))

        # oddball real-world examples
        self.assertParsed(
            ([], ["roc", "ehsan"]),
            parse(
                "Bug 1094764 - Implement AudioContext.suspend and friends.  r=roc,ehsan"
            ),
        )
        self.assertParsed(
            ([], ["bsmedberg", "dbaron"]),
            parse(
                "Bug 380783 - nsStringAPI.h: no equivalent of IsVoid (tell if "
                "string is null), patch by Mook <mook.moz+mozbz@gmail.com>, "
                "r=bsmedberg/dbaron, sr=dbaron, a1.9=bz"
            ),
        )
        self.assertParsed(
            ([], ["hsinyi"]),
            parse(
                "Bug 1181382: move declaration into namespace to resolve conflict. "
                "r=hsinyi. try: -b d -p all -u none -t none"
            ),
        )
        self.assertParsed(
            ([], ["bsmedberg"]),
            parse(
                "Bug 1024110 - Change Aurora's default profile behavior to use "
                "channel-specific profiles. r=bsmedberg f=gavin,markh"
            ),
        )
        self.assertParsed(
            ([], ["gijs"]),
            parse(
                "Bug 1199050 - Round off the corners of browser-extension-panel's "
                "content. ui-r=maritz, r=gijs"
            ),
        )
        self.assertParsed(
            ([], ["billm"]),
            parse(
                "Bug 1197422 - Part 2: [webext] Implement the pageAction API. "
                "r=billm ui-r=bwinton"
            ),
        )

    def test_morph_blocking_reviewers(self):
        def morph(title):
            commits = [Commit(title=title)]
            submit.morph_blocking_reviewers(commits)
            return commits[0].title

        self.assertEqual("stuff", morph("stuff"))
        self.assertEqual("stuff; r=romulus!", morph("stuff; r!romulus"))
        self.assertEqual(
            "stuff; r=romulus!, r=remus!", morph("stuff; r!romulus, r!remus")
        )
        self.assertEqual("stuff; r=romulus!,remus!", morph("stuff; r!romulus,remus"))
        self.assertEqual(
            "stuff; r=romulus!, r=remus", morph("stuff; r!romulus, r=remus")
        )
        self.assertEqual("stuff; r=romulus!", morph("stuff; r!romulus!"))
        self.assertEqual("stuff; r=romulus!", morph("stuff; r=romulus!"))
        self.assertEqual("stuff; r=romulus!.", morph("stuff; r!romulus."))

        self.assertEqual("stuff; r=a!,a.b!", morph("stuff; r!a,a.b"))
        self.assertEqual("stuff; r=a!, r=a.b", morph("stuff; r!a, r=a.b"))
        self.assertEqual("stuff; r=a.b!", morph("stuff; r!a.b!"))
        self.assertEqual("stuff; r=a.b!", morph("stuff; r=a.b!"))
        self.assertEqual("stuff; r=a.b!.", morph("stuff; r!a.b."))

    def test_arc_diff_rev(self):
        parse = helpers.parse_arc_diff_rev

        self.assertEqual(1, parse("Differential Revision: https://example.com/D1"))
        self.assertEqual(22, parse("Differential Revision: https://example.com/D22"))
        self.assertEqual(1, parse("Differential Revision:https://example.com/D1"))
        self.assertEqual(1, parse("Differential Revision:   https://example.com/D1"))
        self.assertEqual(22, parse("   Differential Revision: https://example.com/D22"))
        self.assertEqual(1, parse("Differential Revision: https://example.com/D1  "))
        self.assertEqual(1, parse("\n  Differential Revision: https://example.com/D1"))
        self.assertEqual(1, parse("Differential Revision: https://example.com/D1\n"))
        self.assertIsNone(parse("Differential Revision: https://example.com/D1 1"))
        self.assertEqual(1, parse("Differential Revision: http://example.com/D1"))
        self.assertIsNone(parse("\n  Differential Revision: http://example.com/Q1"))

    def test_arc_reject(self):
        reject = helpers.has_arc_rejections

        self.assertTrue(reject("Summary: blah\nReviewers: blah\n"))
        self.assertTrue(reject("Reviewers: blah\nSummary: blah\n"))
        self.assertTrue(reject("blah\nSummary: blah\n\nReviewers: blah"))
        self.assertTrue(reject("Summary:\n\nReviewers:\n\n"))
        self.assertFalse(reject("Summary: blah"))
        self.assertFalse(reject("Reviewers: blah"))

    def test_commit_title_is_wip(self):
        self.assertFalse(Commit(title="blah").wip_in_commit_title())
        self.assertTrue(Commit(title="WIP: blah").wip_in_commit_title())
        self.assertTrue(Commit(title="WIP blah").wip_in_commit_title())
        self.assertTrue(Commit(title="WIP").wip_in_commit_title())
        self.assertTrue(Commit(title="wip: blah").wip_in_commit_title())
        self.assertTrue(Commit(title="wip blah").wip_in_commit_title())
        self.assertTrue(Commit(title="wip").wip_in_commit_title())
        self.assertFalse(Commit(title="WIPblah").wip_in_commit_title())
        self.assertFalse(Commit(title=" WIP: blah").wip_in_commit_title())
        self.assertFalse(Commit(title=" WIP blah").wip_in_commit_title())
        self.assertFalse(Commit(title=" WIP").wip_in_commit_title())


if __name__ == "__main__":
    unittest.main()



# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

import argparse
import os
import textwrap
from importlib import import_module
from typing import List, Set

from mozphab import commands

from .config import config
from .detect_repository import find_repo_root
from .logger import logger


def should_fallback_to_submit(argv: List[str], commands: Set[str]) -> bool:
    """Return `True` if `moz-phab` should fallback to `submit` command.

    If `moz-phab` is called without a command and from within a repository,
    default to submit.
    """
    # Just running `moz-phab` with no arguments means fallback.
    if not argv:
        return True

    # Don't fallback if help args are passed.
    if any(help_arg in set(argv) for help_arg in {"-h", "--help"}):
        return False

    # Don't fallback if a known command is passed.
    if argv[0] in commands:
        return False

    # Don't fallback when we aren't in a repo.
    avoid_jj_vcs = "--avoid-jj-vcs" in argv
    if not find_repo_root(os.getcwd(), avoid_jj=avoid_jj_vcs):
        return False

    return True


def parse_args(argv: List[str]) -> argparse.Namespace:
    main_parser = argparse.ArgumentParser(add_help=False)
    main_parser.add_argument("--version", action="store_true", help=argparse.SUPPRESS)
    main_parser.add_argument(
        "--trace", "--debug", action="store_true", help=argparse.SUPPRESS
    )
    main_parser.add_argument(
        "--avoid-jj-vcs",
        action="store_true",
        help=(
            "Suppress discovery of Jujutsu repositories via `.jj`. "
            "Useful if you have a co-located Jujutsu repository, "
            "and would like to use Git instead."
        ),
    )
    parser = argparse.ArgumentParser(
        parents=[main_parser],
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent(
            """
            If moz-phab is executed without specifying a command, the 'submit' command
            will be executed.

            For more help on 'submit' and its options run 'moz-phab submit --help'.
            """
        ),
        epilog=textwrap.dedent(
            f"""\
                configuration:
                    {config.filename}

                documentation:
                    https://github.com/mozilla-conduit/review/blob/master/README.md
            """
        ),
    )

    commands_parser = parser.add_subparsers(
        dest="command",
        metavar="COMMAND",
        description="For full command description: moz-phab COMMAND -h",
    )
    commands_parser.required = True

    for command in commands.__all__:
        module = import_module("mozphab.commands.{}".format(command))
        add_parser = getattr(module, "add_parser", None)
        if callable(add_parser):
            add_parser(commands_parser)
            logger.debug("Command added - %s", command)

    help_parser = commands_parser.add_parser("help")
    help_parser.add_argument("command", nargs=argparse.OPTIONAL)
    help_parser.set_defaults(print_help=True)

    fallback = should_fallback_to_submit(argv, set(commands_parser.choices))
    if fallback:
        argv.insert(0, "submit")

    main_args, unknown = main_parser.parse_known_args(argv)

    # map --version to the 'version' command
    if main_args.version:
        unknown = ["version"]

    args = parser.parse_args(unknown)

    # copy across parsed main_args; they are defined in `args`, but not set
    for name, value in vars(main_args).items():
        args.__setattr__(name, value)

    args.fallback = fallback

    # handle the help command here as printing help needs access to the parser
    if hasattr(args, "print_help"):
        help_argv = ["--help"]
        if args.command:
            help_argv.insert(0, args.command)
        # parse_args calls parser.exit() when passed --help
        parser.parse_args(help_argv)

    return args
